from fontTools.misc.transform import Identity
from fontTools.pens.hashPointPen import HashPointPen
import pytest


class _TestGlyph(object):
    width = 500

    def drawPoints(self, pen):
        pen.beginPath(identifier="abc")
        pen.addPoint((0.0, 0.0), "line", False, "start", identifier="0000")
        pen.addPoint((10, 110), "line", False, None, identifier="0001")
        pen.addPoint((50.0, 75.0), None, False, None, identifier="0002")
        pen.addPoint((60.0, 50.0), None, False, None, identifier="0003")
        pen.addPoint((50.0, 0.0), "curve", True, "last", identifier="0004")
        pen.endPath()


class _TestGlyph2(_TestGlyph):
    def drawPoints(self, pen):
        pen.beginPath(identifier="abc")
        pen.addPoint((0.0, 0.0), "line", False, "start", identifier="0000")
        # Minor difference to _TestGlyph() is in the next line:
        pen.addPoint((101, 10), "line", False, None, identifier="0001")
        pen.addPoint((50.0, 75.0), None, False, None, identifier="0002")
        pen.addPoint((60.0, 50.0), None, False, None, identifier="0003")
        pen.addPoint((50.0, 0.0), "curve", True, "last", identifier="0004")
        pen.endPath()


class _TestGlyph3(_TestGlyph):
    def drawPoints(self, pen):
        pen.beginPath(identifier="abc")
        pen.addPoint((0.0, 0.0), "line", False, "start", identifier="0000")
        pen.addPoint((10, 110), "line", False, None, identifier="0001")
        pen.endPath()
        # Same segment, but in a different path:
        pen.beginPath(identifier="pth2")
        pen.addPoint((50.0, 75.0), None, False, None, identifier="0002")
        pen.addPoint((60.0, 50.0), None, False, None, identifier="0003")
        pen.addPoint((50.0, 0.0), "curve", True, "last", identifier="0004")
        pen.endPath()


class _TestGlyph4(_TestGlyph):
    def drawPoints(self, pen):
        pen.beginPath(identifier="abc")
        pen.addPoint((0.0, 0.0), "move", False, "start", identifier="0000")
        pen.addPoint((10, 110), "line", False, None, identifier="0001")
        pen.addPoint((50.0, 75.0), None, False, None, identifier="0002")
        pen.addPoint((60.0, 50.0), None, False, None, identifier="0003")
        pen.addPoint((50.0, 0.0), "curve", True, "last", identifier="0004")
        pen.endPath()


class _TestGlyph5(_TestGlyph):
    def drawPoints(self, pen):
        pen.addComponent("b", Identity)


class HashPointPenTest(object):
    def test_addComponent(self):
        pen = HashPointPen(_TestGlyph().width, {"a": _TestGlyph()})
        pen.addComponent("a", (2, 0, 0, 3, -10, 5))
        assert pen.hash == "w500[l0+0l10+110o50+75o60+50c50+0|(+2+0+0+3-10+5)]"

    def test_NestedComponents(self):
        pen = HashPointPen(
            _TestGlyph().width, {"a": _TestGlyph5(), "b": _TestGlyph()}
        )  # "a" contains "b" as a component
        pen.addComponent("a", (2, 0, 0, 3, -10, 5))

        assert (
            pen.hash
            == "w500[[l0+0l10+110o50+75o60+50c50+0|(+1+0+0+1+0+0)](+2+0+0+3-10+5)]"
        )

    def test_outlineAndComponent(self):
        pen = HashPointPen(_TestGlyph().width, {"a": _TestGlyph()})
        glyph = _TestGlyph()
        glyph.drawPoints(pen)
        pen.addComponent("a", (2, 0, 0, 2, -10, 5))

        assert (
            pen.hash
            == "w500l0+0l10+110o50+75o60+50c50+0|[l0+0l10+110o50+75o60+50c50+0|(+2+0+0+2-10+5)]"
        )

    def test_addComponent_missing_raises(self):
        pen = HashPointPen(_TestGlyph().width, dict())
        with pytest.raises(KeyError) as excinfo:
            pen.addComponent("a", Identity)
        assert excinfo.value.args[0] == "a"

    def test_similarGlyphs(self):
        pen = HashPointPen(_TestGlyph().width)
        glyph = _TestGlyph()
        glyph.drawPoints(pen)

        pen2 = HashPointPen(_TestGlyph2().width)
        glyph = _TestGlyph2()
        glyph.drawPoints(pen2)

        assert pen.hash != pen2.hash

    def test_similarGlyphs2(self):
        pen = HashPointPen(_TestGlyph().width)
        glyph = _TestGlyph()
        glyph.drawPoints(pen)

        pen2 = HashPointPen(_TestGlyph3().width)
        glyph = _TestGlyph3()
        glyph.drawPoints(pen2)

        assert pen.hash != pen2.hash

    def test_similarGlyphs3(self):
        pen = HashPointPen(_TestGlyph().width)
        glyph = _TestGlyph()
        glyph.drawPoints(pen)

        pen2 = HashPointPen(_TestGlyph4().width)
        glyph = _TestGlyph4()
        glyph.drawPoints(pen2)

        assert pen.hash != pen2.hash

    def test_glyphVsComposite(self):
        # If a glyph contains a component, the decomposed glyph should still
        # compare false
        pen = HashPointPen(_TestGlyph().width, {"a": _TestGlyph()})
        pen.addComponent("a", Identity)

        pen2 = HashPointPen(_TestGlyph().width)
        glyph = _TestGlyph()
        glyph.drawPoints(pen2)

        assert pen.hash != pen2.hash



from fontTools.ttLib import TTFont
from fontTools.varLib.hvar import add_HVAR
from io import StringIO
import os
import unittest
import pytest


def test_roundtrip():
    CWD = os.path.abspath(os.path.dirname(__file__))
    DATADIR = os.path.join(CWD, "data")
    ttx_path = os.path.join(DATADIR, "MutatorSans_All_Variable.ttx")
    font = TTFont()
    font.importXML(ttx_path)

    HVAR1 = StringIO()
    font.saveXML(HVAR1, tables=["HVAR"])

    del font["HVAR"]
    add_HVAR(font)

    HVAR2 = StringIO()
    font.saveXML(HVAR2, tables=["HVAR"])

    assert HVAR1.getvalue() == HVAR2.getvalue()



from fontTools.misc.testTools import FakeFont, getXML, parseXML
from fontTools.misc.textTools import deHexStr, hexStr
from fontTools.ttLib import newTable
import unittest


# Example: Format 0 Ligature Caret Table
# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6lcar.html
LCAR_FORMAT_0_DATA = deHexStr(
    "0001 0000 0000 "  #  0: Version=1.0, Format=0
    "0006 0004 0002 "  #  6: LookupFormat=6, UnitSize=4, NUnits=2
    "0008 0001 0000 "  # 12: SearchRange=8, EntrySelector=1, RangeShift=0
    "0001 001E "  # 18: Glyph=1 (f_r), OffsetOfLigCaretEntry=30
    "0003 0022 "  # 22: Glyph=3 (f_f_l), OffsetOfLigCaretEntry=34
    "FFFF 0000 "  # 26: Glyph=<end>, OffsetOfLigCaretEntry=0
    "0001 00DC "  # 30: DivisionPointCount=1, DivisionPoint=[220]
    "0002 00EF 01D8 "  # 34: DivisionPointCount=2, DivisionPoint=[239, 475]
)  # 40: <end>
assert len(LCAR_FORMAT_0_DATA) == 40


LCAR_FORMAT_0_XML = [
    '<Version value="0x00010000"/>',
    '<LigatureCarets Format="0">',
    "  <Carets>",
    '    <Lookup glyph="f_f_l">',
    "      <!-- DivsionPointCount=2 -->",
    '      <DivisionPoint index="0" value="239"/>',
    '      <DivisionPoint index="1" value="472"/>',
    "    </Lookup>",
    '    <Lookup glyph="f_r">',
    "      <!-- DivsionPointCount=1 -->",
    '      <DivisionPoint index="0" value="220"/>',
    "    </Lookup>",
    "  </Carets>",
    "</LigatureCarets>",
]


# Example: Format 1 Ligature Caret Table
# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6lcar.html
LCAR_FORMAT_1_DATA = deHexStr(
    "0001 0000 0001 "  #  0: Version=1.0, Format=1
    "0006 0004 0002 "  #  6: LookupFormat=6, UnitSize=4, NUnits=2
    "0008 0001 0000 "  # 12: SearchRange=8, EntrySelector=1, RangeShift=0
    "0001 001E "  # 18: Glyph=1 (f_r), OffsetOfLigCaretEntry=30
    "0003 0022 "  # 22: Glyph=3 (f_f_l), OffsetOfLigCaretEntry=34
    "FFFF 0000 "  # 26: Glyph=<end>, OffsetOfLigCaretEntry=0
    "0001 0032 "  # 30: DivisionPointCount=1, DivisionPoint=[50]
    "0002 0037 004B "  # 34: DivisionPointCount=2, DivisionPoint=[55, 75]
)  # 40: <end>
assert len(LCAR_FORMAT_1_DATA) == 40


LCAR_FORMAT_1_XML = [
    '<Version value="0x00010000"/>',
    '<LigatureCarets Format="1">',
    "  <Carets>",
    '    <Lookup glyph="f_f_l">',
    "      <!-- DivsionPointCount=2 -->",
    '      <DivisionPoint index="0" value="55"/>',
    '      <DivisionPoint index="1" value="75"/>',
    "    </Lookup>",
    '    <Lookup glyph="f_r">',
    "      <!-- DivsionPointCount=1 -->",
    '      <DivisionPoint index="0" value="50"/>',
    "    </Lookup>",
    "  </Carets>",
    "</LigatureCarets>",
]


class LCARTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None
        cls.font = FakeFont([".notdef", "f_r", "X", "f_f_l"])

    def test_decompile_toXML_format0(self):
        table = newTable("lcar")
        table.decompile(LCAR_FORMAT_0_DATA, self.font)
        self.assertEqual(getXML(table.toXML), LCAR_FORMAT_0_XML)

    def test_compile_fromXML_format0(self):
        table = newTable("lcar")
        for name, attrs, content in parseXML(LCAR_FORMAT_0_XML):
            table.fromXML(name, attrs, content, font=self.font)
        self.assertEqual(hexStr(table.compile(self.font)), hexStr(LCAR_FORMAT_0_DATA))

    def test_decompile_toXML_format1(self):
        table = newTable("lcar")
        table.decompile(LCAR_FORMAT_1_DATA, self.font)
        self.assertEqual(getXML(table.toXML), LCAR_FORMAT_1_XML)

    def test_compile_fromXML_format1(self):
        table = newTable("lcar")
        for name, attrs, content in parseXML(LCAR_FORMAT_1_XML):
            table.fromXML(name, attrs, content, font=self.font)
        self.assertEqual(hexStr(table.compile(self.font)), hexStr(LCAR_FORMAT_1_DATA))


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



#!/usr/bin/env python3
"""Script to add a suffix to all family names in the input font's `name` table,
and to optionally rename the output files with the given suffix.

The current family name substring is searched in the nameIDs 1, 3, 4, 6, 16,
and 21, and if found the suffix is inserted after it; or else the suffix is
appended at the end.
"""
import os
import argparse
import logging
from fontTools.ttLib import TTFont
from fontTools.misc.cliTools import makeOutputFileName


logger = logging.getLogger()

WINDOWS_ENGLISH_IDS = 3, 1, 0x409
MAC_ROMAN_IDS = 1, 0, 0

FAMILY_RELATED_IDS = dict(
    LEGACY_FAMILY=1,
    TRUETYPE_UNIQUE_ID=3,
    FULL_NAME=4,
    POSTSCRIPT_NAME=6,
    PREFERRED_FAMILY=16,
    WWS_FAMILY=21,
)


def get_current_family_name(table):
    family_name_rec = None
    for plat_id, enc_id, lang_id in (WINDOWS_ENGLISH_IDS, MAC_ROMAN_IDS):
        for name_id in (
            FAMILY_RELATED_IDS["PREFERRED_FAMILY"],
            FAMILY_RELATED_IDS["LEGACY_FAMILY"],
        ):
            family_name_rec = table.getName(
                nameID=name_id,
                platformID=plat_id,
                platEncID=enc_id,
                langID=lang_id,
            )
            if family_name_rec is not None:
                break
        if family_name_rec is not None:
            break
    if not family_name_rec:
        raise ValueError("family name not found; can't add suffix")
    return family_name_rec.toUnicode()


def insert_suffix(string, family_name, suffix):
    # check whether family_name is a substring
    start = string.find(family_name)
    if start != -1:
        # insert suffix after the family_name substring
        end = start + len(family_name)
        new_string = string[:end] + suffix + string[end:]
    else:
        # it's not, we just append the suffix at the end
        new_string = string + suffix
    return new_string


def rename_record(name_record, family_name, suffix):
    string = name_record.toUnicode()
    new_string = insert_suffix(string, family_name, suffix)
    name_record.string = new_string
    return string, new_string


def rename_file(filename, family_name, suffix):
    filename, ext = os.path.splitext(filename)
    ps_name = family_name.replace(" ", "")
    if ps_name in filename:
        ps_suffix = suffix.replace(" ", "")
        return insert_suffix(filename, ps_name, ps_suffix) + ext
    else:
        return insert_suffix(filename, family_name, suffix) + ext


def add_family_suffix(font, suffix):
    table = font["name"]

    family_name = get_current_family_name(table)
    logger.info("  Current family name: '%s'", family_name)

    # postcript name can't contain spaces
    ps_family_name = family_name.replace(" ", "")
    ps_suffix = suffix.replace(" ", "")
    for rec in table.names:
        name_id = rec.nameID
        if name_id not in FAMILY_RELATED_IDS.values():
            continue
        if name_id == FAMILY_RELATED_IDS["POSTSCRIPT_NAME"]:
            old, new = rename_record(rec, ps_family_name, ps_suffix)
        elif name_id == FAMILY_RELATED_IDS["TRUETYPE_UNIQUE_ID"]:
            # The Truetype Unique ID rec may contain either the PostScript
            # Name or the Full Name string, so we try both
            if ps_family_name in rec.toUnicode():
                old, new = rename_record(rec, ps_family_name, ps_suffix)
            else:
                old, new = rename_record(rec, family_name, suffix)
        else:
            old, new = rename_record(rec, family_name, suffix)
        logger.info("    %r: '%s' -> '%s'", rec, old, new)

    return family_name


def main(args=None):
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("-s", "--suffix", required=True)
    parser.add_argument("input_fonts", metavar="FONTFILE", nargs="+")
    output_group = parser.add_mutually_exclusive_group()
    output_group.add_argument("-i", "--inplace", action="store_true")
    output_group.add_argument("-d", "--output-dir")
    output_group.add_argument("-o", "--output-file")
    parser.add_argument("-R", "--rename-files", action="store_true")
    parser.add_argument("-v", "--verbose", action="count", default=0)
    options = parser.parse_args(args)

    if not options.verbose:
        level = "WARNING"
    elif options.verbose == 1:
        level = "INFO"
    else:
        level = "DEBUG"
    logging.basicConfig(level=level, format="%(message)s")

    if options.output_file and len(options.input_fonts) > 1:
        parser.error("argument -o/--output-file can't be used with multiple inputs")
    if options.rename_files and (options.inplace or options.output_file):
        parser.error("argument -R not allowed with arguments -i or -o")

    for input_name in options.input_fonts:
        logger.info("Renaming font: '%s'", input_name)

        font = TTFont(input_name)
        family_name = add_family_suffix(font, options.suffix)

        if options.inplace:
            output_name = input_name
        elif options.output_file:
            output_name = options.output_file
        else:
            if options.rename_files:
                input_name = rename_file(input_name, family_name, options.suffix)
            output_name = makeOutputFileName(input_name, options.output_dir)

        font.save(output_name)
        logger.info("Saved font: '%s'", output_name)

        font.close()
        del font

    logger.info("Done!")


if __name__ == "__main__":
    main()



from fontTools.misc.testTools import FakeFont, getXML, parseXML
from fontTools.misc.textTools import deHexStr, hexStr
from fontTools.ttLib import newTable
import unittest


# Example: Format 0 Optical Bounds Table
# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6opbd.html
OPBD_FORMAT_0_DATA = deHexStr(
    "0001 0000 0000 "  #  0: Version=1.0, Format=0
    "0006 0004 0002 "  #  6: LookupFormat=6, UnitSize=4, NUnits=2
    "0008 0001 0000 "  # 12: SearchRange=8, EntrySelector=1, RangeShift=0
    "000A 001E "  # 18: Glyph=10(=C), OffsetOfOpticalBoundsDeltas=30
    "002B 0026 "  # 22: Glyph=43(=A), OffsetOfOpticalBoundsDeltas=38
    "FFFF 0000 "  # 26: Glyph=<end>, OffsetOfOpticalBoundsDeltas=0
    "FFCE 0005 0037 FFFB "  # 30: Bounds[C].Left=-50 .Top=5 .Right=55 .Bottom=-5
    "FFF6 000F 0000 0000 "  # 38: Bounds[A].Left=-10 .Top=15 .Right=0 .Bottom=0
)  # 46: <end>
assert len(OPBD_FORMAT_0_DATA) == 46


OPBD_FORMAT_0_XML = [
    '<Version value="0x00010000"/>',
    '<OpticalBounds Format="0">',
    "  <OpticalBoundsDeltas>",
    '    <Lookup glyph="A">',
    '      <Left value="-10"/>',
    '      <Top value="15"/>',
    '      <Right value="0"/>',
    '      <Bottom value="0"/>',
    "    </Lookup>",
    '    <Lookup glyph="C">',
    '      <Left value="-50"/>',
    '      <Top value="5"/>',
    '      <Right value="55"/>',
    '      <Bottom value="-5"/>',
    "    </Lookup>",
    "  </OpticalBoundsDeltas>",
    "</OpticalBounds>",
]


# Example: Format 1 Optical Bounds Table
# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6opbd.html
OPBD_FORMAT_1_DATA = deHexStr(
    "0001 0000 0001 "  #  0: Version=1.0, Format=1
    "0006 0004 0002 "  #  6: LookupFormat=6, UnitSize=4, NUnits=2
    "0008 0001 0000 "  # 12: SearchRange=8, EntrySelector=1, RangeShift=0
    "000A 001E "  # 18: Glyph=10(=C), OffsetOfOpticalBoundsPoints=30
    "002B 0026 "  # 22: Glyph=43(=A), OffsetOfOpticalBoundsPoints=38
    "FFFF 0000 "  # 26: Glyph=<end>, OffsetOfOpticalBoundsPoints=0
    "0024 0025 0026 0027 "  # 30: Bounds[C].Left=36 .Top=37 .Right=38 .Bottom=39
    "0020 0029 FFFF FFFF "  # 38: Bounds[A].Left=32 .Top=41 .Right=-1 .Bottom=-1
)  # 46: <end>
assert len(OPBD_FORMAT_1_DATA) == 46


OPBD_FORMAT_1_XML = [
    '<Version value="0x00010000"/>',
    '<OpticalBounds Format="1">',
    "  <OpticalBoundsPoints>",
    '    <Lookup glyph="A">',
    '      <Left value="32"/>',
    '      <Top value="41"/>',
    '      <Right value="-1"/>',
    '      <Bottom value="-1"/>',
    "    </Lookup>",
    '    <Lookup glyph="C">',
    '      <Left value="36"/>',
    '      <Top value="37"/>',
    '      <Right value="38"/>',
    '      <Bottom value="39"/>',
    "    </Lookup>",
    "  </OpticalBoundsPoints>",
    "</OpticalBounds>",
]


# This is the content of the Optical Bounds table in AppleChancery.ttf,
# font version 8.0d1e1 of 2013-02-06. An early version of fontTools
# was crashing when trying to decompile this table.
# https://github.com/fonttools/fonttools/issues/1031
OPBD_APPLE_CHANCERY_DATA = deHexStr(
    "0001 0000 0000 "  #   0: Version=1.0, Format=0
    "0004 0006 0011 "  #   6: LookupFormat=4, UnitSize=6, NUnits=17
    "0060 0004 0006 "  #  12: SearchRange=96, EntrySelector=4, RangeShift=6
    "017d 017d 0072 "  #  18: Seg[0].LastGlyph=381, FirstGlyph=381, Off=114(+6)
    "0183 0180 0074 "  #  24: Seg[1].LastGlyph=387, FirstGlyph=384, Off=116(+6)
    "0186 0185 007c "  #  30: Seg[2].LastGlyph=390, FirstGlyph=389, Off=124(+6)
    "018f 018b 0080 "  #  36: Seg[3].LastGlyph=399, FirstGlyph=395, Off=128(+6)
    "01a0 0196 008a "  #  42: Seg[4].LastGlyph=416, FirstGlyph=406, Off=138(+6)
    "01a5 01a3 00a0 "  #  48: Seg[5].LastGlyph=421, FirstGlyph=419, Off=160(+6)
    "01aa 01aa 00a6 "  #  54: Seg[6].LastGlyph=426, FirstGlyph=426, Off=166(+6)
    "01ac 01ac 00a8 "  #  60: Seg[7].LastGlyph=428, FirstGlyph=428, Off=168(+6)
    "01fb 01f1 00aa "  #  66: Seg[8].LastGlyph=507, FirstGlyph=497, Off=170(+6)
    "0214 0209 00c0 "  #  72: Seg[9].LastGlyph=532, FirstGlyph=521, Off=192(+6)
    "021d 0216 00d8 "  #  78: Seg[10].LastGlyph=541, FirstGlyph=534, Off=216(+6)
    "0222 0220 00e8 "  #  84: Seg[11].LastGlyph=546, FirstGlyph=544, Off=232(+6)
    "0227 0225 00ee "  #  90: Seg[12].LastGlyph=551, FirstGlyph=549, Off=238(+6)
    "0229 0229 00f4 "  #  96: Seg[13].LastGlyph=553, FirstGlyph=553, Off=244(+6)
    "023b 023b 00f6 "  # 102: Seg[14].LastGlyph=571, FirstGlyph=571, Off=246(+6)
    "023e 023e 00f8 "  # 108: Seg[15].LastGlyph=574, FirstGlyph=574, Off=248(+6)
    "ffff ffff 00fa "  # 114: Seg[16]=<end>
    "0100 0108 0110 0118 0120 0128 0130 0138 0140 0148 0150 0158 "
    "0160 0168 0170 0178 0180 0188 0190 0198 01a0 01a8 01b0 01b8 "
    "01c0 01c8 01d0 01d8 01e0 01e8 01f0 01f8 0200 0208 0210 0218 "
    "0220 0228 0230 0238 0240 0248 0250 0258 0260 0268 0270 0278 "
    "0280 0288 0290 0298 02a0 02a8 02b0 02b8 02c0 02c8 02d0 02d8 "
    "02e0 02e8 02f0 02f8 0300 0308 0310 0318 fd98 0000 0000 0000 "
    "fdbc 0000 0000 0000 fdbc 0000 0000 0000 fdbf 0000 0000 0000 "
    "fdbc 0000 0000 0000 fd98 0000 0000 0000 fda9 0000 0000 0000 "
    "fd98 0000 0000 0000 fd98 0000 0000 0000 fd98 0000 0000 0000 "
    "0000 0000 0205 0000 0000 0000 0205 0000 0000 0000 02a4 0000 "
    "0000 0000 027e 0000 0000 0000 02f4 0000 0000 0000 02a4 0000 "
    "0000 0000 0365 0000 0000 0000 0291 0000 0000 0000 0291 0000 "
    "0000 0000 026a 0000 0000 0000 02b8 0000 0000 0000 02cb 0000 "
    "0000 0000 02a4 0000 0000 0000 01a9 0000 0000 0000 0244 0000 "
    "0000 0000 02a4 0000 0000 0000 02cb 0000 0000 0000 0244 0000 "
    "0000 0000 0307 0000 0000 0000 0307 0000 0000 0000 037f 0000 "
    "0000 0000 0307 0000 0000 0000 0307 0000 0000 0000 0307 0000 "
    "0000 0000 0307 0000 0000 0000 0307 0000 0000 0000 03e3 0000 "
    "0000 0000 030c 0000 0000 0000 0307 0000 fe30 0000 0000 0000 "
    "fe7e 0000 0000 0000 fe91 0000 0000 0000 fe6a 0000 0000 0000 "
    "fe6a 0000 0000 0000 fecb 0000 0000 0000 fe6a 0000 0000 0000 "
    "fe7e 0000 0000 0000 fea4 0000 0000 0000 fe7e 0000 0000 0000 "
    "fe44 0000 0000 0000 fea4 0000 0000 0000 feb8 0000 0000 0000 "
    "fe7e 0000 0000 0000 fe5e 0000 0000 0000 fe37 0000 0000 0000 "
    "fe37 0000 0000 0000 fcbd 0000 0000 0000 fd84 0000 0000 0000 "
    "fd98 0000 0000 0000 fd82 0000 0000 0000 fcbd 0000 0000 0000 "
    "fd84 0000 0000 0000 fcbd 0000 0000 0000 fcbd 0000 0000 0000 "
    "fe72 0000 0000 0000 ff9d 0000 0000 0000 0000 0000 032f 0000 "
    "0000 0000 03ba 0000 "
)
assert len(OPBD_APPLE_CHANCERY_DATA) == 800


class OPBDTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None
        glyphs = [".notdef"] + ["X.alt%d" for g in range(1, 50)]
        glyphs[10] = "C"
        glyphs[43] = "A"
        cls.font = FakeFont(glyphs)

    def test_decompile_toXML_format0(self):
        table = newTable("opbd")
        table.decompile(OPBD_FORMAT_0_DATA, self.font)
        self.assertEqual(getXML(table.toXML), OPBD_FORMAT_0_XML)

    def test_compile_fromXML_format0(self):
        table = newTable("opbd")
        for name, attrs, content in parseXML(OPBD_FORMAT_0_XML):
            table.fromXML(name, attrs, content, font=self.font)
        self.assertEqual(hexStr(table.compile(self.font)), hexStr(OPBD_FORMAT_0_DATA))

    def test_decompile_toXML_format1(self):
        table = newTable("opbd")
        table.decompile(OPBD_FORMAT_1_DATA, self.font)
        self.assertEqual(getXML(table.toXML), OPBD_FORMAT_1_XML)

    def test_compile_fromXML_format1(self):
        table = newTable("opbd")
        for name, attrs, content in parseXML(OPBD_FORMAT_1_XML):
            table.fromXML(name, attrs, content, font=self.font)
        self.assertEqual(hexStr(table.compile(self.font)), hexStr(OPBD_FORMAT_1_DATA))

    def test_decompile_AppleChancery(self):
        # Make sure we do not crash when decompiling the 'opbd' table of
        # AppleChancery.ttf. https://github.com/fonttools/fonttools/issues/1031
        table = newTable("opbd")
        table.decompile(OPBD_APPLE_CHANCERY_DATA, self.font)
        self.assertIn('<OpticalBounds Format="0">', getXML(table.toXML))


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from .otBase import BaseTTXConverter


# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6gcid.html
class table__g_c_i_d(BaseTTXConverter):
    """Glyph ID to CID table

    The AAT ``gcid`` table stores glyphID-to-CID mappings.

    See also https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6gcid.html
    """

    pass



"""Generic visitor pattern implementation for Python objects."""

import enum


class Visitor(object):
    defaultStop = False

    @classmethod
    def _register(celf, clazzes_attrs):
        assert celf != Visitor, "Subclass Visitor instead."
        if "_visitors" not in celf.__dict__:
            celf._visitors = {}

        def wrapper(method):
            assert method.__name__ == "visit"
            for clazzes, attrs in clazzes_attrs:
                if type(clazzes) != tuple:
                    clazzes = (clazzes,)
                if type(attrs) == str:
                    attrs = (attrs,)
                for clazz in clazzes:
                    _visitors = celf._visitors.setdefault(clazz, {})
                    for attr in attrs:
                        assert attr not in _visitors, (
                            "Oops, class '%s' has visitor function for '%s' defined already."
                            % (clazz.__name__, attr)
                        )
                        _visitors[attr] = method
            return None

        return wrapper

    @classmethod
    def register(celf, clazzes):
        if type(clazzes) != tuple:
            clazzes = (clazzes,)
        return celf._register([(clazzes, (None,))])

    @classmethod
    def register_attr(celf, clazzes, attrs):
        clazzes_attrs = []
        if type(clazzes) != tuple:
            clazzes = (clazzes,)
        if type(attrs) == str:
            attrs = (attrs,)
        for clazz in clazzes:
            clazzes_attrs.append((clazz, attrs))
        return celf._register(clazzes_attrs)

    @classmethod
    def register_attrs(celf, clazzes_attrs):
        return celf._register(clazzes_attrs)

    @classmethod
    def _visitorsFor(celf, thing, _default={}):
        typ = type(thing)

        for celf in celf.mro():
            _visitors = getattr(celf, "_visitors", None)
            if _visitors is None:
                break

            for base in typ.mro():
                m = celf._visitors.get(base, None)
                if m is not None:
                    return m

        return _default

    def visitObject(self, obj, *args, **kwargs):
        """Called to visit an object. This function loops over all non-private
        attributes of the objects and calls any user-registered (via
        @register_attr() or @register_attrs()) visit() functions.

        If there is no user-registered visit function, of if there is and it
        returns True, or it returns None (or doesn't return anything) and
        visitor.defaultStop is False (default), then the visitor will proceed
        to call self.visitAttr()"""

        keys = sorted(vars(obj).keys())
        _visitors = self._visitorsFor(obj)
        defaultVisitor = _visitors.get("*", None)
        for key in keys:
            if key[0] == "_":
                continue
            value = getattr(obj, key)
            visitorFunc = _visitors.get(key, defaultVisitor)
            if visitorFunc is not None:
                ret = visitorFunc(self, obj, key, value, *args, **kwargs)
                if ret == False or (ret is None and self.defaultStop):
                    continue
            self.visitAttr(obj, key, value, *args, **kwargs)

    def visitAttr(self, obj, attr, value, *args, **kwargs):
        """Called to visit an attribute of an object."""
        self.visit(value, *args, **kwargs)

    def visitList(self, obj, *args, **kwargs):
        """Called to visit any value that is a list."""
        for value in obj:
            self.visit(value, *args, **kwargs)

    def visitDict(self, obj, *args, **kwargs):
        """Called to visit any value that is a dictionary."""
        for value in obj.values():
            self.visit(value, *args, **kwargs)

    def visitLeaf(self, obj, *args, **kwargs):
        """Called to visit any value that is not an object, list,
        or dictionary."""
        pass

    def visit(self, obj, *args, **kwargs):
        """This is the main entry to the visitor. The visitor will visit object
        obj.

        The visitor will first determine if there is a registered (via
        @register()) visit function for the type of object. If there is, it
        will be called, and (visitor, obj, *args, **kwargs) will be passed to
        the user visit function.

        If there is no user-registered visit function, of if there is and it
        returns True, or it returns None (or doesn't return anything) and
        visitor.defaultStop is False (default), then the visitor will proceed
        to dispatch to one of self.visitObject(), self.visitList(),
        self.visitDict(), or self.visitLeaf() (any of which can be overriden in
        a subclass)."""

        visitorFunc = self._visitorsFor(obj).get(None, None)
        if visitorFunc is not None:
            ret = visitorFunc(self, obj, *args, **kwargs)
            if ret == False or (ret is None and self.defaultStop):
                return
        if hasattr(obj, "__dict__") and not isinstance(obj, enum.Enum):
            self.visitObject(obj, *args, **kwargs)
        elif isinstance(obj, list):
            self.visitList(obj, *args, **kwargs)
        elif isinstance(obj, dict):
            self.visitDict(obj, *args, **kwargs)
        else:
            self.visitLeaf(obj, *args, **kwargs)



import os

import pytest
import py

from fontTools.qu2cu.cli import _main as main
from fontTools.ttLib import TTFont


DATADIR = os.path.join(os.path.dirname(__file__), "data")

TEST_TTFS = [
    py.path.local(DATADIR).join("NotoSansArabic-Regular.quadratic.subset.ttf"),
]


@pytest.fixture
def test_paths(tmpdir):
    result = []
    for path in TEST_TTFS:
        new_path = tmpdir / path.basename
        path.copy(new_path)
        result.append(new_path)
    return result


class MainTest(object):
    @staticmethod
    def run_main(*args):
        main([str(p) for p in args if p])

    def test_no_output(self, test_paths):
        ttf_path = test_paths[0]

        self.run_main(ttf_path)

        output_path = str(ttf_path).replace(".ttf", ".cubic.ttf")
        font = TTFont(output_path)
        assert font["head"].glyphDataFormat == 1
        assert os.stat(ttf_path).st_size > os.stat(output_path).st_size

    def test_output_file(self, test_paths):
        ttf_path = test_paths[0]
        output_path = str(ttf_path) + ".cubic"

        self.run_main(ttf_path, "-o", output_path)

        font = TTFont(output_path)
        assert font["head"].glyphDataFormat == 1

    def test_stats(self, test_paths):
        ttf_path = test_paths[0]
        self.run_main(ttf_path, "--verbose")

    def test_all_cubic(self, test_paths):
        ttf_path = test_paths[0]

        self.run_main(ttf_path, "-c")

        output_path = str(ttf_path).replace(".ttf", ".cubic.ttf")
        font = TTFont(output_path)
        assert font["head"].glyphDataFormat == 1



import unittest

# Python 3 renamed 'assertRaisesRegexp' to 'assertRaisesRegex', and fires
# deprecation warnings if a program uses the old name.
if not hasattr(unittest.TestCase, "assertRaisesRegex"):
    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp



from fontTools.misc.testTools import getXML, parseXML, parseXmlInto, FakeFont
from fontTools.misc.textTools import deHexStr, hexStr
from fontTools.misc.xmlWriter import XMLWriter
from fontTools.ttLib.tables.otBase import OTTableReader, OTTableWriter
import fontTools.ttLib.tables.otTables as otTables
from io import StringIO
from textwrap import dedent
import unittest


def makeCoverage(glyphs):
    coverage = otTables.Coverage()
    coverage.glyphs = glyphs
    return coverage


class SingleSubstTest(unittest.TestCase):
    def setUp(self):
        self.glyphs = ".notdef A B C D E a b c d e".split()
        self.font = FakeFont(self.glyphs)

    def test_postRead_format1(self):
        table = otTables.SingleSubst()
        table.Format = 1
        rawTable = {"Coverage": makeCoverage(["A", "B", "C"]), "DeltaGlyphID": 5}
        table.postRead(rawTable, self.font)
        self.assertEqual(table.mapping, {"A": "a", "B": "b", "C": "c"})

    def test_postRead_format2(self):
        table = otTables.SingleSubst()
        table.Format = 2
        rawTable = {
            "Coverage": makeCoverage(["A", "B", "C"]),
            "GlyphCount": 3,
            "Substitute": ["c", "b", "a"],
        }
        table.postRead(rawTable, self.font)
        self.assertEqual(table.mapping, {"A": "c", "B": "b", "C": "a"})

    def test_postRead_formatUnknown(self):
        table = otTables.SingleSubst()
        table.Format = 987
        rawTable = {"Coverage": makeCoverage(["A", "B", "C"])}
        self.assertRaises(AssertionError, table.postRead, rawTable, self.font)

    def test_preWrite_format1(self):
        table = otTables.SingleSubst()
        table.mapping = {"A": "a", "B": "b", "C": "c"}
        rawTable = table.preWrite(self.font)
        self.assertEqual(table.Format, 1)
        self.assertEqual(rawTable["Coverage"].glyphs, ["A", "B", "C"])
        self.assertEqual(rawTable["DeltaGlyphID"], 5)

    def test_preWrite_format2(self):
        table = otTables.SingleSubst()
        table.mapping = {"A": "c", "B": "b", "C": "a"}
        rawTable = table.preWrite(self.font)
        self.assertEqual(table.Format, 2)
        self.assertEqual(rawTable["Coverage"].glyphs, ["A", "B", "C"])
        self.assertEqual(rawTable["Substitute"], ["c", "b", "a"])

    def test_preWrite_emptyMapping(self):
        table = otTables.SingleSubst()
        table.mapping = {}
        rawTable = table.preWrite(self.font)
        self.assertEqual(table.Format, 2)
        self.assertEqual(rawTable["Coverage"].glyphs, [])
        self.assertEqual(rawTable["Substitute"], [])

    def test_toXML2(self):
        writer = XMLWriter(StringIO())
        table = otTables.SingleSubst()
        table.mapping = {"A": "a", "B": "b", "C": "c"}
        table.toXML2(writer, self.font)
        self.assertEqual(
            writer.file.getvalue().splitlines()[1:],
            [
                '<Substitution in="A" out="a"/>',
                '<Substitution in="B" out="b"/>',
                '<Substitution in="C" out="c"/>',
            ],
        )

    def test_fromXML(self):
        table = otTables.SingleSubst()
        for name, attrs, content in parseXML(
            '<Substitution in="A" out="a"/>'
            '<Substitution in="B" out="b"/>'
            '<Substitution in="C" out="c"/>'
        ):
            table.fromXML(name, attrs, content, self.font)
        self.assertEqual(table.mapping, {"A": "a", "B": "b", "C": "c"})


class MultipleSubstTest(unittest.TestCase):
    def setUp(self):
        self.glyphs = ".notdef c f i t c_t f_f_i".split()
        self.font = FakeFont(self.glyphs)

    def test_postRead_format1(self):
        makeSequence = otTables.MultipleSubst.makeSequence_
        table = otTables.MultipleSubst()
        table.Format = 1
        rawTable = {
            "Coverage": makeCoverage(["c_t", "f_f_i"]),
            "Sequence": [makeSequence(["c", "t"]), makeSequence(["f", "f", "i"])],
        }
        table.postRead(rawTable, self.font)
        self.assertEqual(table.mapping, {"c_t": ["c", "t"], "f_f_i": ["f", "f", "i"]})

    def test_postRead_formatUnknown(self):
        table = otTables.MultipleSubst()
        table.Format = 987
        self.assertRaises(AssertionError, table.postRead, {}, self.font)

    def test_preWrite_format1(self):
        table = otTables.MultipleSubst()
        table.mapping = {"c_t": ["c", "t"], "f_f_i": ["f", "f", "i"]}
        rawTable = table.preWrite(self.font)
        self.assertEqual(table.Format, 1)
        self.assertEqual(rawTable["Coverage"].glyphs, ["c_t", "f_f_i"])

    def test_toXML2(self):
        writer = XMLWriter(StringIO())
        table = otTables.MultipleSubst()
        table.mapping = {"c_t": ["c", "t"], "f_f_i": ["f", "f", "i"]}
        table.toXML2(writer, self.font)
        self.assertEqual(
            writer.file.getvalue().splitlines()[1:],
            [
                '<Substitution in="c_t" out="c,t"/>',
                '<Substitution in="f_f_i" out="f,f,i"/>',
            ],
        )

    def test_fromXML(self):
        table = otTables.MultipleSubst()
        for name, attrs, content in parseXML(
            '<Substitution in="c_t" out="c,t"/>'
            '<Substitution in="f_f_i" out="f,f,i"/>'
        ):
            table.fromXML(name, attrs, content, self.font)
        self.assertEqual(table.mapping, {"c_t": ["c", "t"], "f_f_i": ["f", "f", "i"]})

    def test_fromXML_oldFormat(self):
        table = otTables.MultipleSubst()
        for name, attrs, content in parseXML(
            "<Coverage>"
            '  <Glyph value="c_t"/>'
            '  <Glyph value="f_f_i"/>'
            "</Coverage>"
            '<Sequence index="0">'
            '  <Substitute index="0" value="c"/>'
            '  <Substitute index="1" value="t"/>'
            "</Sequence>"
            '<Sequence index="1">'
            '  <Substitute index="0" value="f"/>'
            '  <Substitute index="1" value="f"/>'
            '  <Substitute index="2" value="i"/>'
            "</Sequence>"
        ):
            table.fromXML(name, attrs, content, self.font)
        self.assertEqual(table.mapping, {"c_t": ["c", "t"], "f_f_i": ["f", "f", "i"]})

    def test_fromXML_oldFormat_bug385(self):
        # https://github.com/fonttools/fonttools/issues/385
        table = otTables.MultipleSubst()
        table.Format = 1
        for name, attrs, content in parseXML(
            "<Coverage>"
            '  <Glyph value="o"/>'
            '  <Glyph value="l"/>'
            "</Coverage>"
            "<Sequence>"
            '  <Substitute value="o"/>'
            '  <Substitute value="l"/>'
            '  <Substitute value="o"/>'
            "</Sequence>"
            "<Sequence>"
            '  <Substitute value="o"/>'
            "</Sequence>"
        ):
            table.fromXML(name, attrs, content, self.font)
        self.assertEqual(table.mapping, {"o": ["o", "l", "o"], "l": ["o"]})


class LigatureSubstTest(unittest.TestCase):
    def setUp(self):
        self.glyphs = ".notdef c f i t c_t f_f f_i f_f_i".split()
        self.font = FakeFont(self.glyphs)

    def makeLigature(self, s):
        """'ffi' --> Ligature(LigGlyph='f_f_i', Component=['f', 'f', 'i'])"""
        lig = otTables.Ligature()
        lig.Component = list(s)
        lig.LigGlyph = "_".join(lig.Component)
        return lig

    def makeLigatures(self, s):
        """'ffi fi' --> [otTables.Ligature, otTables.Ligature]"""
        return [self.makeLigature(lig) for lig in s.split()]

    def test_postRead_format1(self):
        table = otTables.LigatureSubst()
        table.Format = 1
        ligs_c = otTables.LigatureSet()
        ligs_c.Ligature = self.makeLigatures("ct")
        ligs_f = otTables.LigatureSet()
        ligs_f.Ligature = self.makeLigatures("ffi ff fi")
        rawTable = {
            "Coverage": makeCoverage(["c", "f"]),
            "LigatureSet": [ligs_c, ligs_f],
        }
        table.postRead(rawTable, self.font)
        self.assertEqual(set(table.ligatures.keys()), {"c", "f"})
        self.assertEqual(len(table.ligatures["c"]), 1)
        self.assertEqual(table.ligatures["c"][0].LigGlyph, "c_t")
        self.assertEqual(table.ligatures["c"][0].Component, ["c", "t"])
        self.assertEqual(len(table.ligatures["f"]), 3)
        self.assertEqual(table.ligatures["f"][0].LigGlyph, "f_f_i")
        self.assertEqual(table.ligatures["f"][0].Component, ["f", "f", "i"])
        self.assertEqual(table.ligatures["f"][1].LigGlyph, "f_f")
        self.assertEqual(table.ligatures["f"][1].Component, ["f", "f"])
        self.assertEqual(table.ligatures["f"][2].LigGlyph, "f_i")
        self.assertEqual(table.ligatures["f"][2].Component, ["f", "i"])

    def test_postRead_formatUnknown(self):
        table = otTables.LigatureSubst()
        table.Format = 987
        rawTable = {"Coverage": makeCoverage(["f"])}
        self.assertRaises(AssertionError, table.postRead, rawTable, self.font)

    def test_preWrite_format1(self):
        table = otTables.LigatureSubst()
        table.ligatures = {
            "c": self.makeLigatures("ct"),
            "f": self.makeLigatures("ffi ff fi"),
        }
        rawTable = table.preWrite(self.font)
        self.assertEqual(table.Format, 1)
        self.assertEqual(rawTable["Coverage"].glyphs, ["c", "f"])
        [c, f] = rawTable["LigatureSet"]
        self.assertIsInstance(c, otTables.LigatureSet)
        self.assertIsInstance(f, otTables.LigatureSet)
        [ct] = c.Ligature
        self.assertIsInstance(ct, otTables.Ligature)
        self.assertEqual(ct.LigGlyph, "c_t")
        self.assertEqual(ct.Component, ["c", "t"])
        [ffi, ff, fi] = f.Ligature
        self.assertIsInstance(ffi, otTables.Ligature)
        self.assertEqual(ffi.LigGlyph, "f_f_i")
        self.assertEqual(ffi.Component, ["f", "f", "i"])
        self.assertIsInstance(ff, otTables.Ligature)
        self.assertEqual(ff.LigGlyph, "f_f")
        self.assertEqual(ff.Component, ["f", "f"])
        self.assertIsInstance(fi, otTables.Ligature)
        self.assertEqual(fi.LigGlyph, "f_i")
        self.assertEqual(fi.Component, ["f", "i"])

    def test_toXML2(self):
        writer = XMLWriter(StringIO())
        table = otTables.LigatureSubst()
        table.ligatures = {
            "c": self.makeLigatures("ct"),
            "f": self.makeLigatures("ffi ff fi"),
        }
        table.toXML2(writer, self.font)
        self.assertEqual(
            writer.file.getvalue().splitlines()[1:],
            [
                '<LigatureSet glyph="c">',
                '  <Ligature components="c,t" glyph="c_t"/>',
                "</LigatureSet>",
                '<LigatureSet glyph="f">',
                '  <Ligature components="f,f,i" glyph="f_f_i"/>',
                '  <Ligature components="f,f" glyph="f_f"/>',
                '  <Ligature components="f,i" glyph="f_i"/>',
                "</LigatureSet>",
            ],
        )

    def test_fromXML(self):
        table = otTables.LigatureSubst()
        for name, attrs, content in parseXML(
            '<LigatureSet glyph="f">'
            '  <Ligature components="f,f,i" glyph="f_f_i"/>'
            '  <Ligature components="f,f" glyph="f_f"/>'
            "</LigatureSet>"
        ):
            table.fromXML(name, attrs, content, self.font)
        self.assertEqual(set(table.ligatures.keys()), {"f"})
        [ffi, ff] = table.ligatures["f"]
        self.assertEqual(ffi.LigGlyph, "f_f_i")
        self.assertEqual(ffi.Component, ["f", "f", "i"])
        self.assertEqual(ff.LigGlyph, "f_f")
        self.assertEqual(ff.Component, ["f", "f"])


class AlternateSubstTest(unittest.TestCase):
    def setUp(self):
        self.glyphs = ".notdef G G.alt1 G.alt2 Z Z.fina".split()
        self.font = FakeFont(self.glyphs)

    def makeAlternateSet(self, s):
        result = otTables.AlternateSet()
        result.Alternate = s.split()
        return result

    def test_postRead_format1(self):
        table = otTables.AlternateSubst()
        table.Format = 1
        rawTable = {
            "Coverage": makeCoverage(["G", "Z"]),
            "AlternateSet": [
                self.makeAlternateSet("G.alt2 G.alt1"),
                self.makeAlternateSet("Z.fina"),
            ],
        }
        table.postRead(rawTable, self.font)
        self.assertEqual(table.alternates, {"G": ["G.alt2", "G.alt1"], "Z": ["Z.fina"]})

    def test_postRead_formatUnknown(self):
        table = otTables.AlternateSubst()
        table.Format = 987
        self.assertRaises(AssertionError, table.postRead, {}, self.font)

    def test_preWrite_format1(self):
        table = otTables.AlternateSubst()
        table.alternates = {"G": ["G.alt2", "G.alt1"], "Z": ["Z.fina"]}
        rawTable = table.preWrite(self.font)
        self.assertEqual(table.Format, 1)
        self.assertEqual(rawTable["Coverage"].glyphs, ["G", "Z"])
        [g, z] = rawTable["AlternateSet"]
        self.assertIsInstance(g, otTables.AlternateSet)
        self.assertEqual(g.Alternate, ["G.alt2", "G.alt1"])
        self.assertIsInstance(z, otTables.AlternateSet)
        self.assertEqual(z.Alternate, ["Z.fina"])

    def test_toXML2(self):
        writer = XMLWriter(StringIO())
        table = otTables.AlternateSubst()
        table.alternates = {"G": ["G.alt2", "G.alt1"], "Z": ["Z.fina"]}
        table.toXML2(writer, self.font)
        self.assertEqual(
            writer.file.getvalue().splitlines()[1:],
            [
                '<AlternateSet glyph="G">',
                '  <Alternate glyph="G.alt2"/>',
                '  <Alternate glyph="G.alt1"/>',
                "</AlternateSet>",
                '<AlternateSet glyph="Z">',
                '  <Alternate glyph="Z.fina"/>',
                "</AlternateSet>",
            ],
        )

    def test_fromXML(self):
        table = otTables.AlternateSubst()
        for name, attrs, content in parseXML(
            '<AlternateSet glyph="G">'
            '  <Alternate glyph="G.alt2"/>'
            '  <Alternate glyph="G.alt1"/>'
            "</AlternateSet>"
            '<AlternateSet glyph="Z">'
            '  <Alternate glyph="Z.fina"/>'
            "</AlternateSet>"
        ):
            table.fromXML(name, attrs, content, self.font)
        self.assertEqual(table.alternates, {"G": ["G.alt2", "G.alt1"], "Z": ["Z.fina"]})


class RearrangementMorphActionTest(unittest.TestCase):
    def setUp(self):
        self.font = FakeFont([".notdef", "A", "B", "C"])

    def testCompile(self):
        r = otTables.RearrangementMorphAction()
        r.NewState = 0x1234
        r.MarkFirst = r.DontAdvance = r.MarkLast = True
        r.ReservedFlags, r.Verb = 0x1FF0, 0xD
        writer = OTTableWriter()
        r.compile(writer, self.font, actionIndex=None)
        self.assertEqual(hexStr(writer.getAllData()), "1234fffd")

    def testCompileActions(self):
        act = otTables.RearrangementMorphAction()
        self.assertEqual(act.compileActions(self.font, []), (None, None))

    def testDecompileToXML(self):
        r = otTables.RearrangementMorphAction()
        r.decompile(OTTableReader(deHexStr("1234fffd")), self.font, actionReader=None)
        toXML = lambda w, f: r.toXML(w, f, {"Test": "Foo"}, "Transition")
        self.assertEqual(
            getXML(toXML, self.font),
            [
                '<Transition Test="Foo">',
                '  <NewState value="4660"/>',  # 0x1234 = 4660
                '  <Flags value="MarkFirst,DontAdvance,MarkLast"/>',
                '  <ReservedFlags value="0x1FF0"/>',
                '  <Verb value="13"/><!-- ABxCD ⇒ CDxBA -->',
                "</Transition>",
            ],
        )


class ContextualMorphActionTest(unittest.TestCase):
    def setUp(self):
        self.font = FakeFont([".notdef", "A", "B", "C"])

    def testCompile(self):
        a = otTables.ContextualMorphAction()
        a.NewState = 0x1234
        a.SetMark, a.DontAdvance, a.ReservedFlags = True, True, 0x3117
        a.MarkIndex, a.CurrentIndex = 0xDEAD, 0xBEEF
        writer = OTTableWriter()
        a.compile(writer, self.font, actionIndex=None)
        self.assertEqual(hexStr(writer.getAllData()), "1234f117deadbeef")

    def testCompileActions(self):
        act = otTables.ContextualMorphAction()
        self.assertEqual(act.compileActions(self.font, []), (None, None))

    def testDecompileToXML(self):
        a = otTables.ContextualMorphAction()
        a.decompile(
            OTTableReader(deHexStr("1234f117deadbeef")), self.font, actionReader=None
        )
        toXML = lambda w, f: a.toXML(w, f, {"Test": "Foo"}, "Transition")
        self.assertEqual(
            getXML(toXML, self.font),
            [
                '<Transition Test="Foo">',
                '  <NewState value="4660"/>',  # 0x1234 = 4660
                '  <Flags value="SetMark,DontAdvance"/>',
                '  <ReservedFlags value="0x3117"/>',
                '  <MarkIndex value="57005"/>',  # 0xDEAD = 57005
                '  <CurrentIndex value="48879"/>',  # 0xBEEF = 48879
                "</Transition>",
            ],
        )


class LigatureMorphActionTest(unittest.TestCase):
    def setUp(self):
        self.font = FakeFont([".notdef", "A", "B", "C"])

    def testDecompileToXML(self):
        a = otTables.LigatureMorphAction()
        actionReader = OTTableReader(deHexStr("DEADBEEF 7FFFFFFE 80000003"))
        a.decompile(OTTableReader(deHexStr("1234FAB30001")), self.font, actionReader)
        toXML = lambda w, f: a.toXML(w, f, {"Test": "Foo"}, "Transition")
        self.assertEqual(
            getXML(toXML, self.font),
            [
                '<Transition Test="Foo">',
                '  <NewState value="4660"/>',  # 0x1234 = 4660
                '  <Flags value="SetComponent,DontAdvance"/>',
                '  <ReservedFlags value="0x1AB3"/>',
                '  <Action GlyphIndexDelta="-2" Flags="Store"/>',
                '  <Action GlyphIndexDelta="3"/>',
                "</Transition>",
            ],
        )

    def testCompileActions_empty(self):
        act = otTables.LigatureMorphAction()
        actions, actionIndex = act.compileActions(self.font, [])
        self.assertEqual(actions, b"")
        self.assertEqual(actionIndex, {})

    def testCompileActions_shouldShareSubsequences(self):
        state = otTables.AATState()
        t = state.Transitions = {i: otTables.LigatureMorphAction() for i in range(3)}
        ligs = [otTables.LigAction() for _ in range(3)]
        for i, lig in enumerate(ligs):
            lig.GlyphIndexDelta = i
        t[0].Actions = ligs[1:2]
        t[1].Actions = ligs[0:3]
        t[2].Actions = ligs[1:3]
        actions, actionIndex = t[0].compileActions(self.font, [state])
        self.assertEqual(actions, deHexStr("00000000 00000001 80000002 80000001"))
        self.assertEqual(
            actionIndex,
            {
                deHexStr("00000000 00000001 80000002"): 0,
                deHexStr("00000001 80000002"): 1,
                deHexStr("80000002"): 2,
                deHexStr("80000001"): 3,
            },
        )


class InsertionMorphActionTest(unittest.TestCase):
    MORPH_ACTION_XML = [
        '<Transition Test="Foo">',
        '  <NewState value="4660"/>',  # 0x1234 = 4660
        '  <Flags value="SetMark,DontAdvance,CurrentIsKashidaLike,'
        'MarkedIsKashidaLike,CurrentInsertBefore,MarkedInsertBefore"/>',
        '  <CurrentInsertionAction glyph="B"/>',
        '  <CurrentInsertionAction glyph="C"/>',
        '  <MarkedInsertionAction glyph="B"/>',
        '  <MarkedInsertionAction glyph="A"/>',
        '  <MarkedInsertionAction glyph="D"/>',
        "</Transition>",
    ]

    def setUp(self):
        self.font = FakeFont([".notdef", "A", "B", "C", "D"])
        self.maxDiff = None

    def testDecompileToXML(self):
        a = otTables.InsertionMorphAction()
        actionReader = OTTableReader(
            deHexStr("DEAD BEEF 0002 0001 0004 0002 0003 DEAD BEEF")
        )
        a.decompile(
            OTTableReader(deHexStr("1234 FC43 0005 0002")), self.font, actionReader
        )
        toXML = lambda w, f: a.toXML(w, f, {"Test": "Foo"}, "Transition")
        self.assertEqual(getXML(toXML, self.font), self.MORPH_ACTION_XML)

    def testCompileFromXML(self):
        a = otTables.InsertionMorphAction()
        for name, attrs, content in parseXML(self.MORPH_ACTION_XML):
            a.fromXML(name, attrs, content, self.font)
        writer = OTTableWriter()
        a.compile(
            writer,
            self.font,
            actionIndex={("B", "C"): 9, ("B", "A", "D"): 7},
        )
        self.assertEqual(hexStr(writer.getAllData()), "1234fc4300090007")

    def testCompileActions_empty(self):
        act = otTables.InsertionMorphAction()
        actions, actionIndex = act.compileActions(self.font, [])
        self.assertEqual(actions, b"")
        self.assertEqual(actionIndex, {})

    def testCompileActions_shouldShareSubsequences(self):
        state = otTables.AATState()
        t = state.Transitions = {i: otTables.InsertionMorphAction() for i in range(3)}
        t[1].CurrentInsertionAction = []
        t[0].MarkedInsertionAction = ["A"]
        t[1].CurrentInsertionAction = ["C", "D"]
        t[1].MarkedInsertionAction = ["B"]
        t[2].CurrentInsertionAction = ["B", "C", "D"]
        t[2].MarkedInsertionAction = ["C", "D"]
        actions, actionIndex = t[0].compileActions(self.font, [state])
        self.assertEqual(actions, deHexStr("0002 0003 0004 0001"))
        self.assertEqual(
            actionIndex,
            {
                ("A",): 3,
                ("B",): 0,
                ("B", "C"): 0,
                ("B", "C", "D"): 0,
                ("C",): 1,
                ("C", "D"): 1,
                ("D",): 2,
            },
        )


class SplitMultipleSubstTest:
    def overflow(self, itemName, itemRecord):
        from fontTools.otlLib.builder import buildMultipleSubstSubtable
        from fontTools.ttLib.tables.otBase import OverflowErrorRecord

        oldSubTable = buildMultipleSubstSubtable(
            {"e": 1, "a": 2, "b": 3, "c": 4, "d": 5}
        )
        newSubTable = otTables.MultipleSubst()

        ok = otTables.splitMultipleSubst(
            oldSubTable,
            newSubTable,
            OverflowErrorRecord((None, None, None, itemName, itemRecord)),
        )

        assert ok
        return oldSubTable.mapping, newSubTable.mapping

    def test_Coverage(self):
        oldMapping, newMapping = self.overflow("Coverage", None)
        assert oldMapping == {"a": 2, "b": 3}
        assert newMapping == {"c": 4, "d": 5, "e": 1}

    def test_RangeRecord(self):
        oldMapping, newMapping = self.overflow("RangeRecord", None)
        assert oldMapping == {"a": 2, "b": 3}
        assert newMapping == {"c": 4, "d": 5, "e": 1}

    def test_Sequence(self):
        oldMapping, newMapping = self.overflow("Sequence", 4)
        assert oldMapping == {"a": 2, "b": 3, "c": 4}
        assert newMapping == {"d": 5, "e": 1}


def test_splitMarkBasePos():
    from fontTools.otlLib.builder import buildAnchor, buildMarkBasePosSubtable

    marks = {
        "acutecomb": (0, buildAnchor(0, 600)),
        "gravecomb": (0, buildAnchor(0, 590)),
        "cedillacomb": (1, buildAnchor(0, 0)),
    }
    bases = {
        "a": {
            0: buildAnchor(350, 500),
            1: None,
        },
        "c": {
            0: buildAnchor(300, 700),
            1: buildAnchor(300, 0),
        },
    }
    glyphOrder = ["a", "c", "acutecomb", "gravecomb", "cedillacomb"]
    glyphMap = {g: i for i, g in enumerate(glyphOrder)}

    oldSubTable = buildMarkBasePosSubtable(marks, bases, glyphMap)
    newSubTable = otTables.MarkBasePos()

    ok = otTables.splitMarkBasePos(oldSubTable, newSubTable, overflowRecord=None)

    assert ok

    assert getXML(oldSubTable.toXML) == [
        '<MarkBasePos Format="1">',
        "  <MarkCoverage>",
        '    <Glyph value="acutecomb"/>',
        '    <Glyph value="gravecomb"/>',
        "  </MarkCoverage>",
        "  <BaseCoverage>",
        '    <Glyph value="a"/>',
        '    <Glyph value="c"/>',
        "  </BaseCoverage>",
        "  <!-- ClassCount=1 -->",
        "  <MarkArray>",
        "    <!-- MarkCount=2 -->",
        '    <MarkRecord index="0">',
        '      <Class value="0"/>',
        '      <MarkAnchor Format="1">',
        '        <XCoordinate value="0"/>',
        '        <YCoordinate value="600"/>',
        "      </MarkAnchor>",
        "    </MarkRecord>",
        '    <MarkRecord index="1">',
        '      <Class value="0"/>',
        '      <MarkAnchor Format="1">',
        '        <XCoordinate value="0"/>',
        '        <YCoordinate value="590"/>',
        "      </MarkAnchor>",
        "    </MarkRecord>",
        "  </MarkArray>",
        "  <BaseArray>",
        "    <!-- BaseCount=2 -->",
        '    <BaseRecord index="0">',
        '      <BaseAnchor index="0" Format="1">',
        '        <XCoordinate value="350"/>',
        '        <YCoordinate value="500"/>',
        "      </BaseAnchor>",
        "    </BaseRecord>",
        '    <BaseRecord index="1">',
        '      <BaseAnchor index="0" Format="1">',
        '        <XCoordinate value="300"/>',
        '        <YCoordinate value="700"/>',
        "      </BaseAnchor>",
        "    </BaseRecord>",
        "  </BaseArray>",
        "</MarkBasePos>",
    ]

    assert getXML(newSubTable.toXML) == [
        '<MarkBasePos Format="1">',
        "  <MarkCoverage>",
        '    <Glyph value="cedillacomb"/>',
        "  </MarkCoverage>",
        "  <BaseCoverage>",
        '    <Glyph value="a"/>',
        '    <Glyph value="c"/>',
        "  </BaseCoverage>",
        "  <!-- ClassCount=1 -->",
        "  <MarkArray>",
        "    <!-- MarkCount=1 -->",
        '    <MarkRecord index="0">',
        '      <Class value="0"/>',
        '      <MarkAnchor Format="1">',
        '        <XCoordinate value="0"/>',
        '        <YCoordinate value="0"/>',
        "      </MarkAnchor>",
        "    </MarkRecord>",
        "  </MarkArray>",
        "  <BaseArray>",
        "    <!-- BaseCount=2 -->",
        '    <BaseRecord index="0">',
        '      <BaseAnchor index="0" empty="1"/>',
        "    </BaseRecord>",
        '    <BaseRecord index="1">',
        '      <BaseAnchor index="0" Format="1">',
        '        <XCoordinate value="300"/>',
        '        <YCoordinate value="0"/>',
        "      </BaseAnchor>",
        "    </BaseRecord>",
        "  </BaseArray>",
        "</MarkBasePos>",
    ]


class ColrV1Test(unittest.TestCase):
    def setUp(self):
        self.font = FakeFont([".notdef", "meh"])

    def test_traverseEmptyPaintColrLayersNeedsNoLayerList(self):
        colr = parseXmlInto(
            self.font,
            otTables.COLR(),
            """
          <Version value="1"/>
          <BaseGlyphList>
            <BaseGlyphPaintRecord index="0">
              <BaseGlyph value="meh"/>
              <Paint Format="1"><!-- PaintColrLayers -->
                <NumLayers value="0"/>
                <FirstLayerIndex value="42"/>
              </Paint>
            </BaseGlyphPaintRecord>
          </BaseGlyphList>
          """,
        )
        paint = colr.BaseGlyphList.BaseGlyphPaintRecord[0].Paint

        # Just want to confirm we don't crash
        visited = []
        paint.traverse(colr, lambda p: visited.append(p))
        assert len(visited) == 1


def test_parse_Device_DeltaValue_from_XML_and_compile():
    # https://github.com/fonttools/fonttools/pull/3757
    font = FakeFont([".notdef", "five"])

    gpos_xml = dedent(
        """\
        <Version value="0x00010000"/>
        <ScriptList>
          <!-- ScriptCount=1 -->
          <ScriptRecord index="0">
            <ScriptTag value="DFLT"/>
            <Script>
              <DefaultLangSys>
                <ReqFeatureIndex value="65535"/>
                <!-- FeatureCount=1 -->
                <FeatureIndex index="0" value="0"/>
              </DefaultLangSys>
              <!-- LangSysCount=0 -->
            </Script>
          </ScriptRecord>
        </ScriptList>
        <FeatureList>
          <!-- FeatureCount=1 -->
          <FeatureRecord index="0">
            <FeatureTag value="curs"/>
            <Feature>
              <!-- LookupCount=1 -->
              <LookupListIndex index="0" value="0"/>
            </Feature>
          </FeatureRecord>
        </FeatureList>
        <LookupList>
          <!-- LookupCount=1 -->
          <Lookup index="0">
            <LookupType value="3"/>
            <LookupFlag value="0"/>
            <!-- SubTableCount=1 -->
            <CursivePos index="0" Format="1">
              <Coverage>
                <Glyph value="five"/>
              </Coverage>
              <!-- EntryExitCount=1 -->
              <EntryExitRecord index="0">
                <EntryAnchor Format="3">
                  <XCoordinate value="124"/>
                  <YCoordinate value="-4"/>
                  <XDeviceTable>
                    <StartSize value="8"/>
                    <EndSize value="9"/>
                    <DeltaFormat value="2"/>
                    <DeltaValue value="[1, 2]"/>
                  </XDeviceTable>
                  <YDeviceTable>
                    <StartSize value="7"/>
                    <EndSize value="7"/>
                    <DeltaFormat value="2"/>
                    <DeltaValue value="[3]"/>
                  </YDeviceTable>
                </EntryAnchor>
                <ExitAnchor Format="2">
                  <XCoordinate value="3"/>
                  <YCoordinate value="4"/>
                  <AnchorPoint value="2"/>
                </ExitAnchor>
              </EntryExitRecord>
            </CursivePos>
          </Lookup>
        </LookupList>"""
    )

    gpos = parseXmlInto(font, otTables.GPOS(), gpos_xml)

    anchor = gpos.LookupList.Lookup[0].SubTable[0].EntryExitRecord[0].EntryAnchor
    assert anchor.XDeviceTable.DeltaValue == [1, 2]
    assert anchor.YDeviceTable.DeltaValue == [3]

    writer = OTTableWriter()
    gpos.compile(writer, font)
    data = writer.getAllData()

    reader = OTTableReader(data, tableTag="GPOS")
    gpos2 = otTables.GPOS()
    gpos2.decompile(reader, font)

    assert dedent("\n".join(getXML(gpos2.toXML, font)[1:-1])) == gpos_xml


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())
